@@ -79,63 +79,86 @@ static int parse_prelogin_xml(struct openconnect_info *vpninfo, xmlNode *xml_nod
 	struct oc_auth_form *form = NULL;
 	struct oc_form_opt *opt, *opt2;
 	char *prompt = NULL, *username_label = NULL, *password_label = NULL;
-	char *saml_method = NULL, *saml_path = NULL;
+	char *s = NULL, *saml_method = NULL, *saml_path = NULL;
 	int result = 0;
 
 	if (!xmlnode_is_named(xml_node, "prelogin-response"))
 		goto out;
 
 	for (xml_node = xml_node->children; xml_node; xml_node = xml_node->next) {
-		char *s = NULL;
-		if (!xmlnode_get_val(xml_node, "saml-request", &s)) {
-			int len;
-			free(saml_path);
-			saml_path = openconnect_base64_decode(&len, s);
-			if (len < 0) {
-				vpn_progress(vpninfo, PRG_ERR, "Could not decode SAML request as base64: %s\n", s);
-				free(s);
-				result = -EINVAL;
-				goto out;
-			}
-			free(s);
-			realloc_inplace(saml_path, len+1);
-			if (!saml_path) {
-				result = -ENOMEM;
-				goto out;
-			}
-			saml_path[len] = '\0';
-		} else {
-			xmlnode_get_val(xml_node, "saml-auth-method", &saml_method);
-			xmlnode_get_val(xml_node, "authentication-message", &prompt);
-			xmlnode_get_val(xml_node, "username-label", &username_label);
-			xmlnode_get_val(xml_node, "password-label", &password_label);
-			/* XX: should we save the certificate username from <ccusername/> ? */
-		}
+		xmlnode_get_val(xml_node, "saml-request", &s);
+		xmlnode_get_val(xml_node, "saml-auth-method", &saml_method);
+		xmlnode_get_val(xml_node, "authentication-message", &prompt);
+		xmlnode_get_val(xml_node, "username-label", &username_label);
+		xmlnode_get_val(xml_node, "password-label", &password_label);
+		/* XX: should we save the certificate username from <ccusername/> ? */
 	}
 
-	/* XX: Alt-secret form field must be specified for SAML, because we can't autodetect it */
-	if (saml_method || saml_path) {
-		if (ctx->portal_userauthcookie)
+	if (saml_method && s) {
+		/* Allow the legacy workflow (no GUI setting up open_webview) to keep working */
+		if (!vpninfo->open_webview && ctx->portal_userauthcookie)
 			vpn_progress(vpninfo, PRG_DEBUG, _("SAML authentication required; using portal-userauthcookie to continue SAML.\n"));
-		else if (ctx->portal_prelogonuserauthcookie)
+		else if (!vpninfo->open_webview && ctx->portal_prelogonuserauthcookie)
 			vpn_progress(vpninfo, PRG_DEBUG, _("SAML authentication required; using portal-prelogonuserauthcookie to continue SAML.\n"));
-		else if (ctx->alt_secret)
+		else if (!vpninfo->open_webview && ctx->alt_secret)
 			vpn_progress(vpninfo, PRG_DEBUG, _("Destination form field %s was specified; assuming SAML %s authentication is complete.\n"),
 			             ctx->alt_secret, saml_method);
 		else {
-			if (saml_method && !strcmp(saml_method, "REDIRECT"))
-				vpn_progress(vpninfo, PRG_ERR,
-				             _("SAML %s authentication is required via %s\n"),
-				             saml_method, saml_path);
-			else
-				vpn_progress(vpninfo, PRG_ERR,
-				             _("SAML %s authentication is required via external script.\n"),
-				             saml_method);
-			vpn_progress(vpninfo, PRG_ERR,
-			             _("When SAML authentication is complete, specify destination form field by appending :field_name to login URL.\n"));
-			/* XX: EINVAL will lead to "failure to parse response", with unnecessary/confusing extra logging output */
-			result = -EPERM;
-			goto out;
+			if (!strcmp(saml_method, "REDIRECT")) {
+				int len;
+				saml_path = openconnect_base64_decode(&len, s);
+				if (len < 0) {
+					vpn_progress(vpninfo, PRG_ERR, "Could not decode SAML request as base64: %s\n", s);
+					free(s);
+					result = -EINVAL;
+					goto out;
+				}
+				free(s);
+				realloc_inplace(saml_path, len+1);
+				if (!saml_path) {
+					result = -ENOMEM;
+					goto out;
+				}
+				saml_path[len] = '\0';
+				vpninfo->sso_login = strdup(saml_path);
+				prompt = strdup("SAML REDIRECT authentication in progress");
+				if (!vpninfo->sso_login || !prompt) {
+					result = -ENOMEM;
+					goto out;
+				}
+			} else if (!strcmp(saml_method, "POST")) {
+				const char *prefix = "data:text/html;base64,";
+				saml_path = s;
+				realloc_inplace(saml_path, strlen(saml_path)+strlen(prefix)+1);
+				if (!saml_path) {
+					result = -ENOMEM;
+					goto out;
+				}
+				memmove(saml_path + strlen(prefix), saml_path, strlen(saml_path) + 1);
+				memcpy(saml_path, prefix, strlen(prefix));
+				vpninfo->sso_login = strdup(saml_path);
+				prompt = strdup("SAML REDIRECT authentication in progress");
+				if (!vpninfo->sso_login || !prompt) {
+					result = -ENOMEM;
+					goto out;
+				}
+			} else {
+				vpn_progress(vpninfo, PRG_ERR, "Unknown SAML method %s\n", saml_method);
+				result = -EINVAL;
+				goto out;
+			}
+
+			vpn_progress(vpninfo, PRG_INFO,
+					_("SAML %s authentication is required via %s\n"),
+					saml_method, saml_path);
+
+			/* Legacy flow (when not called by n-m-oc) */
+			if (!vpninfo->open_webview) {
+				vpn_progress(vpninfo,
+					PRG_ERR, _("When SAML authentication is complete, specify destination form field by appending :field_name to login URL.\n"));
+				result = -EINVAL;
+				goto out;
+			}
 		}
 	}
 
@@ -156,10 +179,12 @@ static int parse_prelogin_xml(struct openconnect_info *vpninfo, xmlNode *xml_nod
 	if (!opt)
 		goto nomem;
 	opt->name = strdup("user");
+	if (!opt->name)
+		goto nomem;
 	if (asprintf(&opt->label, "%s: ", username_label ? : _("Username")) == 0)
 		goto nomem;
 	if (!ctx->username)
-		opt->type = OC_FORM_OPT_TEXT;
+		opt->type = saml_path ? OC_FORM_OPT_SSO_USER : OC_FORM_OPT_TEXT;
 	else {
 		opt->type = OC_FORM_OPT_HIDDEN;
 		opt->_value = ctx->username;
@@ -171,6 +196,8 @@ static int parse_prelogin_xml(struct openconnect_info *vpninfo, xmlNode *xml_nod
 	if (!opt2)
 		goto nomem;
 	opt2->name = strdup(ctx->alt_secret ? : "passwd");
+	if (!opt2->name)
+		goto nomem;
 	if (asprintf(&opt2->label, "%s: ", ctx->alt_secret ? : password_label ? : _("Password")) == 0)
 		goto nomem;
 
@@ -182,16 +209,18 @@ static int parse_prelogin_xml(struct openconnect_info *vpninfo, xmlNode *xml_nod
 	 * password in the first form means we should treat the first
 	 * form's password as a token field.
 	 */
-	if (!can_gen_tokencode(vpninfo, form, opt2) && !ctx->alt_secret
-	    && password_label && strcmp(password_label, "Password"))
+	if (saml_path)
+		opt2->type = OC_FORM_OPT_SSO_TOKEN;
+	else if (!can_gen_tokencode(vpninfo, form, opt2) && !ctx->alt_secret
+	         && password_label && strcmp(password_label, "Password"))
 		opt2->type = OC_FORM_OPT_TOKEN;
 	else
 		opt2->type = OC_FORM_OPT_PASSWORD;
 
 	vpn_progress(vpninfo, PRG_TRACE, "Prelogin form %s: \"%s\" %s(%s)=%s, \"%s\" %s(%s)\n",
 	             form->auth_id,
-	             opt->label, opt->name, opt->type == OC_FORM_OPT_TEXT ? "TEXT" : "HIDDEN", opt->_value,
-	             opt2->label, opt2->name, opt2->type == OC_FORM_OPT_PASSWORD ? "PASSWORD" : "TOKEN");
+	             opt->label, opt->name, opt->type == OC_FORM_OPT_SSO_USER ? "SSO" : opt->type == OC_FORM_OPT_TEXT ? "TEXT" : "HIDDEN", opt->_value,
+	             opt2->label, opt2->name, opt2->type == OC_FORM_OPT_SSO_TOKEN ? "SSO" : opt2->type == OC_FORM_OPT_PASSWORD ? "PASSWORD" : "TOKEN");
 
 out:
 	free(prompt);
@@ -371,8 +400,9 @@ static int parse_login_xml(struct openconnect_info *vpninfo, xmlNode *xml_node,
 
 	if (unknown_args)
 		vpn_progress(vpninfo, PRG_ERR,
-					 _("Please report %d unexpected values above (of which %d fatal) to <openconnect-devel@lists.infradead.org>\n"),
-					 unknown_args, fatal_args);
+					 _("Please report %d unexpected values above (of which %d fatal) to <%s>\n"),
+					 unknown_args, fatal_args,
+					 "openconnect-devel@lists.infradead.org");
 	if (fatal_args) {
 		buf_free(cookie);
 		return -EPERM;
@@ -575,7 +605,6 @@ static int gpst_login(struct openconnect_info *vpninfo, int portal, struct login
 {
 	int result, blind_retry = 0;
 	struct oc_text_buf *request_body = buf_alloc();
-	const char *request_body_type = "application/x-www-form-urlencoded";
 	char *xml_buf = NULL, *orig_path;
 
 	/* Ask the user to fill in the auth form; repeat as necessary */
@@ -613,6 +642,15 @@ static int gpst_login(struct openconnect_info *vpninfo, int portal, struct login
 		if (result)
 			goto out;
 
+		/* Coming back from SAML we might have been redirected */
+		if (vpninfo->redirect_url) {
+			result = handle_redirect(vpninfo);
+			free(vpninfo->redirect_url);
+			vpninfo->redirect_url = NULL;
+			if (result)
+				goto out;
+		}
+
 	replay_form:
 		/* generate token code if specified */
 		result = do_gen_tokencode(vpninfo, ctx->form);
@@ -624,7 +662,7 @@ static int gpst_login(struct openconnect_info *vpninfo, int portal, struct login
 
 		/* submit gateway login (ssl-vpn/login.esp) or portal config (global-protect/getconfig.esp) request */
 		buf_truncate(request_body);
-		buf_append(request_body, "jnlpReady=jnlpReady&ok=Login&direct=yes&clientVer=4100&prot=https:");
+		buf_append(request_body, "jnlpReady=jnlpReady&ok=Login&direct=yes&clientVer=4100&prot=https:&internal=no");
 		append_opt(request_body, "ipv6-support", vpninfo->disable_ipv6 ? "no" : "yes");
 		append_opt(request_body, "clientos", gpst_os_name(vpninfo));
 		append_opt(request_body, "os-version", vpninfo->platname);
@@ -647,7 +685,7 @@ static int gpst_login(struct openconnect_info *vpninfo, int portal, struct login
 
 		orig_path = vpninfo->urlpath;
 		vpninfo->urlpath = strdup(portal ? "global-protect/getconfig.esp" : "ssl-vpn/login.esp");
-		result = do_https_request(vpninfo, "POST", request_body_type, request_body, &xml_buf, NULL, HTTP_NO_FLAGS);
+		result = do_https_request(vpninfo, "POST", "application/x-www-form-urlencoded", request_body, &xml_buf, NULL, HTTP_NO_FLAGS);
 		free(vpninfo->urlpath);
 		vpninfo->urlpath = orig_path;
 
@@ -737,8 +775,6 @@ int gpst_bye(struct openconnect_info *vpninfo, const char *reason)
 	char *orig_path;
 	int result;
 	struct oc_text_buf *request_body = buf_alloc();
-	const char *request_body_type = "application/x-www-form-urlencoded";
-	const char *method = "POST";
 	char *xml_buf = NULL;
 
 	/* In order to logout successfully, the client must send not only
@@ -763,7 +799,7 @@ int gpst_bye(struct openconnect_info *vpninfo, const char *reason)
 	orig_path = vpninfo->urlpath;
 	vpninfo->urlpath = strdup("ssl-vpn/logout.esp");
 	openconnect_close_https(vpninfo, 0);
-	result = do_https_request(vpninfo, method, request_body_type, request_body, &xml_buf, NULL, HTTP_NO_FLAGS);
+	result = do_https_request(vpninfo, "POST", "application/x-www-form-urlencoded", request_body, &xml_buf, NULL, HTTP_NO_FLAGS);
 	free(vpninfo->urlpath);
 	vpninfo->urlpath = orig_path;
 
