@@ -18,31 +18,34 @@
 #ifndef __OPENCONNECT_GNUTLS_H__
 #define __OPENCONNECT_GNUTLS_H__
 
+#include "openconnect-internal.h"
+
 #include <gnutls/gnutls.h>
 #include <gnutls/pkcs12.h>
 #include <gnutls/abstract.h>
 
-#include "openconnect-internal.h"
+int load_tpm1_key(struct openconnect_info *vpninfo, struct cert_info *certinfo,
+		  gnutls_datum_t *fdata, gnutls_privkey_t *pkey, gnutls_datum_t *pkey_sig);
+void release_tpm1_ctx(struct openconnect_info *info, struct cert_info *certinfo);
 
-int load_tpm1_key(struct openconnect_info *vpninfo, gnutls_datum_t *fdata,
-		  gnutls_privkey_t *pkey, gnutls_datum_t *pkey_sig);
-void release_tpm1_ctx(struct openconnect_info *info);
-
-int load_tpm2_key(struct openconnect_info *vpninfo, gnutls_datum_t *fdata,
-		 gnutls_privkey_t *pkey, gnutls_datum_t *pkey_sig);
-void release_tpm2_ctx(struct openconnect_info *info);
-int install_tpm2_key(struct openconnect_info *vpninfo, gnutls_privkey_t *pkey, gnutls_datum_t *pkey_sig,
+int load_tpm2_key(struct openconnect_info *vpninfo, struct cert_info *certinfo,
+		  gnutls_datum_t *fdata, gnutls_privkey_t *pkey, gnutls_datum_t *pkey_sig);
+void release_tpm2_ctx(struct openconnect_info *info, struct cert_info *certinfo);
+int install_tpm2_key(struct openconnect_info *vpninfo, struct cert_info *certinfo,
+		     gnutls_privkey_t *pkey, gnutls_datum_t *pkey_sig,
 		     unsigned int parent, int emptyauth, int legacy,
 		     gnutls_datum_t *privdata, gnutls_datum_t *pubdata);
 
 int tpm2_rsa_sign_hash_fn(gnutls_privkey_t key, gnutls_sign_algorithm_t algo,
-			  void *_vpninfo, unsigned int flags,
+			  void *_certinfo, unsigned int flags,
 			  const gnutls_datum_t *data, gnutls_datum_t *sig);
 int tpm2_ec_sign_hash_fn(gnutls_privkey_t key, gnutls_sign_algorithm_t algo,
-			 void *_vpninfo, unsigned int flags,
+			 void *_certinfo, unsigned int flags,
 			 const gnutls_datum_t *data, gnutls_datum_t *sig);
-int oc_pkcs1_pad(struct openconnect_info *vpninfo,
-		 unsigned char *buf, int size, const gnutls_datum_t *data);
+int oc_pad_rsasig(struct openconnect_info *vpninfo, gnutls_sign_algorithm_t algo,
+		  unsigned char *buf, int size, const gnutls_datum_t *data, int keybits);
+uint16_t tpm2_key_curve(struct openconnect_info *vpninfo, struct cert_info *certinfo);
+int tpm2_rsa_key_bits(struct openconnect_info *vpninfo, struct cert_info *certinfo);
 
 /* GnuTLS 3.6.0+ provides this. We have our own for older GnuTLS. There is
  * also _gnutls_encode_ber_rs_raw() in some older versions, but there were
@@ -61,7 +64,7 @@ char *get_gnutls_cipher(gnutls_session_t session);
  * were built again, but we might be run against a version which is
  * newer. So some ancient compatibility code *can* be dropped at
  * compile time. Likewise, if building against GnuTLS 2.x then we
- * can never be running agsinst a 3.x library — the soname changed.
+ * can never be running against a 3.x library — the soname changed.
  *
  * This macro was added upstream, gnutls_check_version_numeric,
  * in 3.5.0 (see https://gitlab.com/gnutls/gnutls/commit/c8b40aeb) */
