@@ -135,6 +135,7 @@ public abstract class LibOpenConnect {
 	public synchronized native void setXMLSHA1(String hash);
 	public synchronized native void setHostname(String hostname);
 	public synchronized native void setVersionString(String version);
+	public synchronized native void setUseragent(String useragent);
 	public synchronized native void setUrlpath(String urlpath);
 	public synchronized native void setLocalName(String localName);
 	public synchronized native void setCAFile(String caFile);
@@ -145,7 +146,6 @@ public abstract class LibOpenConnect {
 	public synchronized native int setTokenMode(int tokenMode, String tokenString);
 	public synchronized native void setCSDWrapper(String wrapper, String TMPDIR, String PATH);
 	public synchronized native void setXMLPost(boolean isEnabled);
-	public synchronized native void setClientCert(String cert, String sslKey);
 	public synchronized native void setReqMTU(int mtu);
 	public synchronized native void setPFS(boolean isEnabled);
 	public synchronized native int setAllowInsecureCrypto(boolean isEnabled);
@@ -155,6 +155,13 @@ public abstract class LibOpenConnect {
 	public synchronized native int disableIPv6();
 	public synchronized native void setCookie(String cookie);
 
+	/* client certificate(s) and private keys */
+
+	public synchronized native int setClientCert(String cert, String sslKey);
+	public synchronized native int setKeyPassword(String password);
+	public synchronized native int setMCACert(String cert, String sslKey);
+	public synchronized native int setMCAKeyPassword(String password);
+
 	/* connection info */
 
 	public synchronized native String getConnectUrl();
@@ -173,7 +180,7 @@ public abstract class LibOpenConnect {
 	public synchronized native int getIdleTimeout();
 	public synchronized native Instant getAuthExpiration();
 
-	/* certificate info */
+	/* peer certificate info */
 
 	public synchronized native int checkPeerCertHash(String hash);
 	public synchronized native String getPeerCertHash();
