@@ -837,21 +837,44 @@ err:
 }
 
 /* special handling: two string arguments */
-JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setClientCert(
+JNIEXPORT int JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setClientCert(
 	JNIEnv *jenv, jobject jobj, jstring jcert, jstring jsslkey)
 {
+	int ret;
 	struct libctx *ctx = getctx(jenv, jobj);
 	const char *cert = NULL, *sslkey = NULL;
 
 	if (!ctx)
-		return;
+		return -EINVAL;
 
-	if (!get_cstring(ctx->jenv, jcert, &cert) &&
-	    !get_cstring(ctx->jenv, jsslkey, &sslkey))
-		openconnect_set_client_cert(ctx->vpninfo, cert, sslkey);
+	if (get_cstring(ctx->jenv, jcert, &cert) ||
+	    get_cstring(ctx->jenv, jsslkey, &sslkey))
+		return -ENOMEM;
+	ret = openconnect_set_client_cert(ctx->vpninfo, cert, sslkey);
 
 	release_cstring(ctx->jenv, jcert, cert);
 	release_cstring(ctx->jenv, jsslkey, sslkey);
+	return ret;
+}
+
+JNIEXPORT int JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setMCACert(
+	JNIEnv *jenv, jobject jobj, jstring jcert, jstring jsslkey)
+{
+	int ret;
+	struct libctx *ctx = getctx(jenv, jobj);
+	const char *cert = NULL, *sslkey = NULL;
+
+	if (!ctx)
+		return -EINVAL;
+
+	if (get_cstring(ctx->jenv, jcert, &cert) ||
+	    get_cstring(ctx->jenv, jsslkey, &sslkey))
+		return -ENOMEM;
+	ret = openconnect_set_mca_cert(ctx->vpninfo, cert, sslkey);
+
+	release_cstring(ctx->jenv, jcert, cert);
+	release_cstring(ctx->jenv, jsslkey, sslkey);
+	return ret;
 }
 
 /* special handling: multiple string arguments */
@@ -1404,6 +1427,14 @@ JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setVersi
 	SET_STRING_END();
 }
 
+JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setUsergent(
+	JNIEnv *jenv, jobject jobj, jstring jarg)
+{
+	SET_STRING_START_VOID()
+	openconnect_set_useragent(ctx->vpninfo, arg);
+	SET_STRING_END();
+}
+
 JNIEXPORT void JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setUrlpath(
 	JNIEnv *jenv, jobject jobj, jstring jarg)
 {
@@ -1580,3 +1611,25 @@ err:
 	openconnect_free_supported_protocols(protos);
 	return NULL;
 }
+
+JNIEXPORT jint JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setKeyPassword(
+	JNIEnv *jenv, jobject jobj, jstring jarg, jstring jarg2)
+{
+	int ret;
+	SET_STRING_START();
+	ret = openconnect_set_key_password(ctx->vpninfo, arg);
+	SET_STRING_END();
+
+	return ret;
+}
+
+JNIEXPORT jint JNICALL Java_org_infradead_libopenconnect_LibOpenConnect_setMCAKeyPassword(
+	JNIEnv *jenv, jobject jobj, jstring jarg, jstring jarg2)
+{
+	int ret;
+	SET_STRING_START();
+	ret = openconnect_set_mca_key_password(ctx->vpninfo, arg);
+	SET_STRING_END();
+
+	return ret;
+}
