@@ -127,6 +127,14 @@
 #include <string.h>
 #include <errno.h>
 
+#ifdef HAVE_POSIX_SPAWN
+#ifdef __APPLE__
+#include <crt_externs.h>
+#define environ (*_NSGetEnviron())
+#endif
+#include <spawn.h>
+#endif
+
 /* Equivalent of "/dev/null" on Windows.
  * See https://stackoverflow.com/a/44163934
  */
@@ -182,7 +190,6 @@ struct pkt {
 			uint32_t spi;
 			uint32_t seq;
 			unsigned char iv[16];
-			unsigned char payload[];
 		} esp;
 		struct {
 			unsigned char pad[2];
@@ -240,7 +247,7 @@ struct pkt {
 #define DTLS_CONNECTED	5	/* Transport connected but not yet enabled */
 #define DTLS_ESTABLISHED 6	/* Data path fully established */
 
-/* Not to be confused with OC_PROTO_xxx flags which are library-visible */
+/* Not to be confused with MULTICERT_PROTO_xxx flags which are library-visible */
 #define PROTO_ANYCONNECT	0
 #define PROTO_NC		1
 #define PROTO_GPST		2
@@ -387,6 +394,8 @@ struct cert_info {
 	char *cert;
 	char *key;
 	char *password;
+
+	void *priv_info;
 #if defined(OPENCONNECT_GNUTLS) && defined(HAVE_TROUSERS)
 	struct oc_tpm1_ctx *tpm1;
 #endif
@@ -488,7 +497,7 @@ struct openconnect_info {
 
 	int cert_expire_warning;
 
-	struct cert_info certinfo[1];
+	struct cert_info certinfo[2];
 
 	char *cafile;
 	unsigned no_system_trust;
@@ -574,13 +583,21 @@ struct openconnect_info {
 	SSL_CTX *https_ctx;
 	SSL *https_ssl;
 	BIO_METHOD *ttls_bio_meth;
+	EC_KEY *strap_key;
+	EC_KEY *strap_dh_key;
 #elif defined(OPENCONNECT_GNUTLS)
 	gnutls_session_t https_sess;
 	gnutls_session_t eap_ttls_sess;
 	gnutls_certificate_credentials_t https_cred;
 	gnutls_psk_client_credentials_t psk_cred;
 	char local_cert_md5[MD5_SIZE * 2 + 1]; /* For CSD */
+	gnutls_privkey_t strap_key;
+	gnutls_privkey_t strap_dh_key;
+	unsigned char finished[64];
+	int finished_len;
 #endif /* OPENCONNECT_GNUTLS */
+	char *strap_pubkey;
+	char *strap_dh_pubkey;
 	char *ciphersuite_config;
 	struct oc_text_buf *ttls_pushbuf;
 	uint8_t ttls_eap_ident;
@@ -762,10 +779,20 @@ struct openconnect_info {
 		DELAY_CLOSE_IMMEDIATE_CALLBACK,
 	} delay_close;                          /* Delay close of mainloop */
 
+	char *sso_login;
+	char *sso_login_final;
+	char *sso_username;
+	char *sso_token_cookie;
+	char *sso_error_cookie;
+	char *sso_cookie_value;
+	char *sso_browser_mode;
+
 	int verbose;
 	void *cbdata;
 	openconnect_validate_peer_cert_vfn validate_peer_cert;
 	openconnect_write_new_config_vfn write_new_config;
+	openconnect_open_webview_vfn open_webview;
+	openconnect_open_webview_vfn open_ext_browser;
 	openconnect_process_auth_form_vfn process_auth_form;
 	openconnect_progress_vfn progress;
 	openconnect_protect_socket_vfn protect_socket;
@@ -791,6 +818,9 @@ struct vpn_proto {
 	/* This does the full authentication, calling back as appropriate */
 	int (*obtain_cookie)(struct openconnect_info *vpninfo);
 
+	/* This checks if SSO authentication is complete */
+	int (*sso_detect_done)(struct openconnect_info *vpninfo, const struct oc_webview_result *result);
+
 	/* Establish the TCP connection (and obtain configuration) */
 	int (*tcp_connect)(struct openconnect_info *vpninfo);
 
@@ -881,8 +911,8 @@ static inline void free_pkt(struct openconnect_info *vpninfo, struct pkt *pkt)
 		free(pkt);
 }
 
-#define vpn_progress(_v, lvl, ...) do {					\
-	if ((_v)->verbose >= (lvl))					\
+#define vpn_progress(_v, lvl, ...) do {				\
+	if ((_v)->verbose >= (lvl))				\
 		(_v)->progress((_v)->cbdata, lvl, __VA_ARGS__);	\
 	} while(0)
 #define vpn_perror(vpninfo, msg) vpn_progress((vpninfo), PRG_ERR, "%s: %s\n", (msg), strerror(errno))
@@ -1144,6 +1174,39 @@ int dumb_socketpair(OPENCONNECT_CMD_SOCKET socks[2], int make_overlapped);
     } while (0)
 
 /****************************************************************************/
+typedef enum {
+	MULTICERT_COMPAT = (1<<0),
+} cert_flag_t;
+
+typedef enum {
+	CERT_FORMAT_ASN1 = 0,
+	CERT_FORMAT_PEM = 1,
+} cert_format_t;
+
+typedef enum {
+	OPENCONNECT_HASH_UNKNOWN = 0,
+#define OPENCONNECT_HASH_NONE OPENCONNECT_HASH_NONE
+	OPENCONNECT_HASH_SHA256 = 1,
+#define OPENCONNECT_HASH_SHA256 OPENCONNECT_HASH_SHA256
+	OPENCONNECT_HASH_SHA384 = 2,
+#define OPENCONNECT_HASH_SHA384 OPENCONNECT_HASH_SHA384
+	OPENCONNECT_HASH_SHA512 = 3,
+#define OPENCONNECT_HASH_SHA512 OPENCONNECT_HASH_SHA512
+	OPENCONNECT_HASH_MAX = OPENCONNECT_HASH_SHA512
+} openconnect_hash_type;
+
+int load_certificate(struct openconnect_info *, struct cert_info *, int flags);
+void unload_certificate(struct cert_info *, int final);
+int export_certificate_pkcs7(struct openconnect_info *, struct cert_info *, cert_format_t format, struct oc_text_buf **);
+
+/* multiple certificate authentication */
+#define MULTICERT_HASH_FLAG(v)	((v)?(1<<((v)-1)):0)
+
+int multicert_sign_data(struct openconnect_info *, struct cert_info *certinfo, unsigned int hashes,
+			const void *data, size_t datalen, struct oc_text_buf **signature);
+
+const char *multicert_hash_get_name(int id);
+openconnect_hash_type multicert_hash_get_id(const char *name);
 
 /* iconv.c */
 #ifdef HAVE_ICONV
@@ -1225,6 +1288,7 @@ int cstp_bye(struct openconnect_info *vpninfo, const char *reason);
 int decompress_and_queue_packet(struct openconnect_info *vpninfo, int compr_type,
 				unsigned char *buf, int len);
 int compress_packet(struct openconnect_info *vpninfo, int compr_type, struct pkt *this);
+int cstp_sso_detect_done(struct openconnect_info *vpninfo, const struct oc_webview_result *result);
 
 /* auth-html.c */
 xmlNodePtr htmlnode_next(xmlNodePtr top, xmlNodePtr node);
@@ -1314,13 +1378,14 @@ int gpst_setup(struct openconnect_info *vpninfo);
 int gpst_mainloop(struct openconnect_info *vpninfo, int *timeout, int readable);
 int gpst_esp_send_probes(struct openconnect_info *vpninfo);
 int gpst_esp_catch_probe(struct openconnect_info *vpninfo, struct pkt *pkt);
+int gpst_sso_detect_done(struct openconnect_info *vpninfo, const struct oc_webview_result *result);
 
 /* lzs.c */
 int lzs_decompress(unsigned char *dst, int dstlen, const unsigned char *src, int srclen);
 int lzs_compress(unsigned char *dst, int dstlen, const unsigned char *src, int srclen);
 
 /* ssl.c */
-unsigned string_is_hostname(const char* str);
+unsigned string_is_hostname(const char *str);
 int connect_https_socket(struct openconnect_info *vpninfo);
 int __attribute__ ((format(printf, 4, 5)))
     request_passphrase(struct openconnect_info *vpninfo, const char *label,
@@ -1354,9 +1419,10 @@ int cancellable_gets(struct openconnect_info *vpninfo, int fd,
 		     char *buf, size_t len);
 
 int cancellable_send(struct openconnect_info *vpninfo, int fd,
-		     char *buf, size_t len);
+		     const char *buf, size_t len);
 int cancellable_recv(struct openconnect_info *vpninfo, int fd,
 		     char *buf, size_t len);
+int cancellable_accept(struct openconnect_info *vpninfo, int fd);
 
 #if defined(OPENCONNECT_OPENSSL)
 /* openssl-pkcs11.c */
@@ -1412,6 +1478,18 @@ int hotp_hmac(struct openconnect_info *vpninfo, const void *challenge);
 int openconnect_install_ctx_verify(struct openconnect_info *vpninfo,
 				   SSL_CTX *ctx);
 #endif
+void free_strap_keys(struct openconnect_info *vpninfo);
+int generate_strap_keys(struct openconnect_info *vpninfo);
+int ecdh_compute_secp256r1(struct openconnect_info *vpninfo, const unsigned char *pubkey,
+			   int pubkey_len, unsigned char *secret);
+int hkdf_sha256_extract_expand(struct openconnect_info *vpninfo, unsigned char *buf,
+			       const char *info, int infolen);
+int aes_256_gcm_decrypt(struct openconnect_info *vpninfo, unsigned char *key,
+			unsigned char *data, int len,
+			unsigned char *iv, unsigned char *tag);
+void append_strap_verify(struct openconnect_info *vpninfo, struct oc_text_buf *buf, int rekey);
+void append_strap_privkey(struct openconnect_info *vpninfo, struct oc_text_buf *buf);
+int ingest_strap_privkey(struct openconnect_info *vpninfo, unsigned char *der, int len);
 
 /* mainloop.c */
 int tun_mainloop(struct openconnect_info *vpninfo, int *timeout, int readable, int did_work);
@@ -1507,8 +1585,8 @@ void buf_append_from_utf16le(struct oc_text_buf *buf, const void *utf16);
 void buf_append_base64(struct oc_text_buf *buf, const void *bytes, int len, int line_len);
 
 /* http.c */
-void dump_buf(struct openconnect_info *vpninfo, char prefix, char *buf);
-void dump_buf_hex(struct openconnect_info *vpninfo, int loglevel, char prefix, unsigned char *buf, int len);
+void do_dump_buf(struct openconnect_info *vpninfo, char prefix, char *buf);
+void do_dump_buf_hex(struct openconnect_info *vpninfo, int loglevel, char prefix, unsigned char *buf, int len);
 char *openconnect_create_useragent(const char *base);
 int process_proxy(struct openconnect_info *vpninfo, int ssl_sock);
 int internal_parse_url(const char *url, char **res_proto, char **res_host,
@@ -1519,6 +1597,7 @@ int do_https_request(struct openconnect_info *vpninfo, const char *method, const
 		     int (*header_cb)(struct openconnect_info *, char *, char *), int flags);
 int http_add_cookie(struct openconnect_info *vpninfo, const char *option,
 		    const char *value, int replace);
+const char *http_get_cookie(struct openconnect_info *vpninfo, const char *name);
 int internal_split_cookies(struct openconnect_info *vpninfo, int replace, const char *def_cookie);
 int urldecode_inplace(char *p);
 int process_http_response(struct openconnect_info *vpninfo, int connect,
@@ -1526,6 +1605,16 @@ int process_http_response(struct openconnect_info *vpninfo, int connect,
 			  struct oc_text_buf *body);
 int handle_redirect(struct openconnect_info *vpninfo);
 void http_common_headers(struct openconnect_info *vpninfo, struct oc_text_buf *buf);
+#define dump_buf(vpninfo, prefix, buf) do {			\
+		if ((vpninfo)->verbose >= PRG_DEBUG) {		\
+			do_dump_buf(vpninfo, prefix, buf);	\
+		}						\
+	} while(0)
+#define dump_buf_hex(vpninfo, loglevel, prefix, buf, len) do {			\
+		if ((vpninfo)->verbose >= (loglevel)) {				\
+			do_dump_buf_hex(vpninfo, loglevel, prefix, buf, len);	\
+		}								\
+	} while(0)
 
 /* http-auth.c */
 void *openconnect_base64_decode(int *len, const char *in);
@@ -1560,8 +1649,11 @@ int process_auth_form(struct openconnect_info *vpninfo, struct oc_auth_form *for
 /* This is private for now since we haven't yet worked out what the API will be */
 void openconnect_set_juniper(struct openconnect_info *vpninfo);
 
+/* hpke.c */
+int handle_external_browser(struct openconnect_info *vpninfo);
+
 /* version.c */
-extern const char *openconnect_version_str;
+extern const char openconnect_version_str[];
 
 
 static inline int certinfo_is_primary(struct cert_info *certinfo)
