@@ -20,9 +20,9 @@
 #ifndef __OPENCONNECT_H__
 #define __OPENCONNECT_H__
 
-#include <stdint.h>
 #include <sys/types.h>
-#include <unistd.h>
+
+#include <stdint.h>
 
 #ifdef __cplusplus
 extern "C" {
@@ -33,9 +33,36 @@ extern "C" {
 #endif
 
 #define OPENCONNECT_API_VERSION_MAJOR 5
-#define OPENCONNECT_API_VERSION_MINOR 4
+#define OPENCONNECT_API_VERSION_MINOR 8
 
 /*
+ * API version 5.8 (v9.00; 2022-04-29):
+ *  - Add openconnect_set_useragent()
+ *  - Add openconnect_set_external_browser_callback()
+ *  - Add openconnect_set_mca_cert() and openconnect_set_mca_key_password()
+ *
+ * API version 5.7 (v8.20; 2022-02-20):
+ *  - Add openconnect_get_connect_url()
+ *  - Add openconnect_set_cookie()
+ *  - Add openconnect_set_allow_insecure_crypto()
+ *  - Add openconnect_get_auth_expiration()
+ *  - Add openconnect_disable_dtls()
+ *  - Make openconnect_disable_ipv6() return int
+ *  - Add openconnect_set_webview_callback()
+ *  - Add openconnect_webview_load_changed()
+ *
+ * API version 5.6 (v8.06; 2020-03-31):
+ *  - Add openconnect_set_trojan_interval()
+ *
+ * API version 5.5 (v8.00; 2019-01-05):
+ *  - add openconnect_set_version_string()
+ *  - add openconnect_set_key_password()
+ *  - Add openconnect_has_tss2_blob_support()
+ *  - Add openconnect_get_supported_protocols()
+ *  - Add openconnect_free_supported_protocols()
+ *  - Add openconnect_get_protocol()
+ *  - Add openconnect_get_idle_timeout()
+ *
  * API version 5.4 (v7.08; 2016-12-13):
  *  - Add openconnect_set_pass_tos()
  *
@@ -166,6 +193,26 @@ extern "C" {
 
 /****************************************************************************/
 
+/* Enumeration of supported VPN protocols */
+
+#define OC_PROTO_PROXY	(1<<0)
+#define OC_PROTO_CSD	(1<<1)
+#define OC_PROTO_AUTH_CERT	(1<<2)
+#define OC_PROTO_AUTH_OTP	(1<<3)
+#define OC_PROTO_AUTH_STOKEN	(1<<4)
+#define OC_PROTO_PERIODIC_TROJAN	(1<<5)
+#define OC_PROTO_HIDDEN	(1<<6)
+#define OC_PROTO_AUTH_MCA	(1<<7)
+
+struct oc_vpn_proto {
+	const char *name;
+	const char *pretty_name;
+	const char *description;
+	unsigned int flags;
+};
+
+/****************************************************************************/
+
 /* Authentication form processing */
 
 #define OC_FORM_OPT_TEXT	1
@@ -173,6 +220,8 @@ extern "C" {
 #define OC_FORM_OPT_SELECT	3
 #define OC_FORM_OPT_HIDDEN	4
 #define OC_FORM_OPT_TOKEN	5
+#define OC_FORM_OPT_SSO_TOKEN	6
+#define OC_FORM_OPT_SSO_USER	7
 
 #define OC_FORM_RESULT_ERR		-1
 #define OC_FORM_RESULT_OK		0
@@ -246,9 +295,10 @@ struct oc_split_include {
 
 struct oc_ip_info {
 	const char *addr;
-	const char *netmask;
+	const char *netmask; /* Just the netmask, in dotted-quad form. */
 	const char *addr6;
-	const char *netmask6;
+	const char *netmask6; /* This is the IPv6 address *and* netmask
+			       * e.g. "2001::dead:beef/128". */
 	const char *dns[3];
 	const char *nbns[3];
 	const char *domain;
@@ -284,6 +334,17 @@ struct oc_cert {
 	void *reserved;
 };
 
+/* Used by openconnect_webview_load_changed() to return data to OC.
+ * The arrays must contain an even number of const strings, with names and
+ * values, and a NULL to terminate. E.g.:
+ * name0, val0, name1, val1, NULL
+ */
+struct oc_webview_result {
+	const char *uri;
+	const char **cookies;
+	const char **headers;
+};
+
 /****************************************************************************/
 
 #define PRG_ERR		0
@@ -319,6 +380,7 @@ typedef enum {
 	OC_TOKEN_MODE_TOTP,
 	OC_TOKEN_MODE_HOTP,
 	OC_TOKEN_MODE_YUBIOATH,
+	OC_TOKEN_MODE_OIDC,
 } oc_token_mode_t;
 
 typedef enum {
@@ -351,7 +413,7 @@ int openconnect_set_csd_environ(struct openconnect_info *vpninfo,
 /* This string is static, valid only while the connection lasts. If you
  * are going to cache this to remember which certs the user has accepted,
  * make sure you also store the host/port for which it was accepted and
- * don't just accept this cert from *anywhere*. Also use use the check
+ * don't just accept this cert from *anywhere*. Also use the check
  * function below instead of manually comparing. When this function
  * returns a string which *doesn't* match the previously-stored hash
  * matched with openconnect_check_peer_cert_hash(), you should store
@@ -396,6 +458,14 @@ int openconnect_set_proxy_auth(struct openconnect_info *vpninfo,
 			       const char *methods);
 int openconnect_set_http_proxy(struct openconnect_info *vpninfo,
 			       const char *proxy);
+
+/* Passing a useragent string to openconnect_vpninfo_new() will still
+ * append the OpenConnect version number. This function allows the
+ * full string to be set, for those cases where a VPN server might
+ * require an *exact* match. */
+int openconnect_set_useragent(struct openconnect_info *vpninfo,
+			      const char *useragent);
+
 int openconnect_passphrase_from_fsid(struct openconnect_info *vpninfo);
 int openconnect_obtain_cookie(struct openconnect_info *vpninfo);
 int openconnect_init_ssl(void);
@@ -408,11 +478,85 @@ int openconnect_init_ssl(void);
 const char *openconnect_get_cstp_cipher(struct openconnect_info *);
 const char *openconnect_get_dtls_cipher(struct openconnect_info *);
 
-/* These return a descriptive string of the compression algorithm 
+/* These return a descriptive string of the compression algorithm
  * in use (LZS, LZ4, ...). If no compression then NULL is returned. */
 const char *openconnect_get_cstp_compression(struct openconnect_info *);
 const char *openconnect_get_dtls_compression(struct openconnect_info *);
 
+
+/*
+ * Since authentication can run in a separate environment to the connection
+ * itself, there is a simple set of information which needs to be passed
+ * from one to the other. Basically it's just the server to connect to, and
+ * the cookie we need for authentication. And luckily, as we've added more
+ * and more protocols over the years, the "cookie" part has remained true
+ * and we haven't needed to use client certificates for the *connection*.
+ *
+ * The *server* part is a little more complex. Firstly, the certificate
+ * might not be valid and may have been accepted manually by the user,
+ * so we pass the certificate fingerprint separately, as a second piece
+ * of information.
+
+ * In the beginning, we passed the server hostname as the third piece of
+ * information, and all was well.
+ *
+ * Then we found servers on a non-standard port, so the authentication
+ * dialogs would use openconnect_get_port() and just append it (":%d")
+ * to the hostname string they passed on.
+ *
+ * Then we encountered servers with round-robin or geo-DNS, which gave
+ * different IP addresses for a given hostname, and we switched the
+ * openconnect_get_hostname() function to return the *IP* address instead,
+ * since the actual name didn't matter when it wasn't being used to check
+ * the server's certificate anyway.
+ *
+ * At some point later, openconnect_get_dnsname() was added to return the
+ * actual hostname, during the authentication phase where the cert was
+ * being presented to the user for manual acceptance. But that wasn't
+ * really important for the authentication → connection handoff.
+ *
+ * Later still, Legacy IP addresses got scarce and SNI was invented, and
+ * we started to see servers behind proxies that forward a connection
+ * based on the SNI in the incoming ClientHello (TLS layer), or based on
+ * the 'Host' header in the initial connection-phase request (HTTP[S]
+ * layer). So returning just the IP address from openconnect_get_hostname()
+ * now made things break.
+ *
+ * So... we need to pass *both* the actual hostname *and* the IP address
+ * to the connecting openconnect invocation. As well as the port.
+ *
+ * In addition, the Pulse protocol introduced a new requirement for the
+ * connection. Instead of connecting to a fixed endpoint on the server,
+ * we must connect to the appropriate *path*, which varies. So in fact
+ * it isn't just the "hostname" any more, but the full URL.
+ *
+ * So, now we have openconnect_get_connect_url() which gets the full URL
+ * including the port and path, and the original hostname.
+ *
+ * Since we're now back to giving openconnect a hostname, we need to add
+ * a '--resolve' argument to avoid the round robin DNS problem and ensure
+ * that we actually connect to the same server we authenticated to. The
+ * arguments for that can be obtained from openconnect_get_dnsname() and
+ * openconnect_get_hostname() — the latter of which, as noted, was changed
+ * years ago to return a numeric address. We end up invoking openconnect
+ * to make the connection as follows:
+ *
+ * openconnect $CONNECT_URL --servercert $FINGERPRINT --cookie $COOKIE \
+ *             --resolve $DNSNAME:$HOSTNAME
+ *
+ * ... where '$HOSTNAME', as returned by openconnect_get_hostname(),
+ * isn't actually a hostname (as noted above). Sorry.
+ *
+ * In fact, what you get back from openconnect_get_hostname() is the
+ * IP literal in the form it would appear in a URL. So IPv6 addresses
+ * are wrapped in [], and that needs to be *stripped* in order to pass
+ * it to openconnect's --resolve argument. As I realise and type this,
+ * it doesn't seem particularly useful to provide yet another function
+ * that will return the non-[]-wrapped version, as we'd still need UI
+ * tools to do it themselves for backward compatibility. Sorry again :)
+ */
+const char *openconnect_get_connect_url(struct openconnect_info *);
+
 /* Returns the IP address of the exact host to which the connection
  * was made. In --cookieonly mode or in any other scenario involving
  * a "two stage" connection, it is important to reconnect by IP because
@@ -483,15 +627,29 @@ void openconnect_set_xmlpost(struct openconnect_info *, int enable);
    trojan binary. */
 int openconnect_set_reported_os(struct openconnect_info *, const char *os);
 
+int openconnect_set_version_string(struct openconnect_info *vpninfo,
+				   const char *version_string);
 int openconnect_set_mobile_info(struct openconnect_info *vpninfo,
 				const char *mobile_platform_version,
 				const char *mobile_device_type,
 				const char *mobile_device_uniqueid);
 int openconnect_set_client_cert(struct openconnect_info *, const char *cert,
 				const char *sslkey);
+int openconnect_set_key_password(struct openconnect_info *vpninfo, const char *pass);
+/**
+ * Multiple certificate authentication (MCA): the client cert _and_ the
+ * mca_cert are used for authentication. The mca_cert is used to sign a
+ * challenge sent by the server.
+ */
+int openconnect_set_mca_cert(struct openconnect_info *, const char *cert,
+			     const char *key);
+int openconnect_set_mca_key_password(struct openconnect_info *vpninfo, const char *pass);
 const char *openconnect_get_ifname(struct openconnect_info *);
 void openconnect_set_reqmtu(struct openconnect_info *, int reqmtu);
 void openconnect_set_dpd(struct openconnect_info *, int min_seconds);
+void openconnect_set_trojan_interval(struct openconnect_info *, int seconds);
+int openconnect_get_idle_timeout(struct openconnect_info *);
+time_t openconnect_get_auth_expiration(struct openconnect_info *);
 
 /* The returned structures are owned by the library and may be freed/replaced
    due to rekey or reconnect. Assume that once the mainloop starts, the
@@ -504,14 +662,17 @@ int openconnect_get_ip_info(struct openconnect_info *,
 
 int openconnect_get_port(struct openconnect_info *);
 const char *openconnect_get_cookie(struct openconnect_info *);
+int openconnect_set_cookie(struct openconnect_info *, const char *);
 void openconnect_clear_cookie(struct openconnect_info *);
 
-void openconnect_disable_ipv6(struct openconnect_info *vpninfo);
+int openconnect_disable_ipv6(struct openconnect_info *vpninfo);
+int openconnect_disable_dtls(struct openconnect_info *vpninfo);
 void openconnect_reset_ssl(struct openconnect_info *vpninfo);
 int openconnect_parse_url(struct openconnect_info *vpninfo, const char *url);
 void openconnect_set_cert_expiry_warning(struct openconnect_info *vpninfo,
 					 int seconds);
 void openconnect_set_pfs(struct openconnect_info *vpninfo, unsigned val);
+int openconnect_set_allow_insecure_crypto(struct openconnect_info *vpninfo, unsigned val);
 
 /* If this is set, then openconnect_obtain_cookie() will abort and return
    failure if the file descriptor is readable. Typically a user may create
@@ -608,6 +769,19 @@ struct openconnect_info *openconnect_vpninfo_new(const char *useragent,
 						 void *privdata);
 void openconnect_vpninfo_free(struct openconnect_info *vpninfo);
 
+typedef int (*openconnect_open_webview_vfn) (struct openconnect_info *,
+					     const char *uri,
+					     void *privdata);
+
+void openconnect_set_webview_callback(struct openconnect_info *vpninfo,
+				      openconnect_open_webview_vfn);
+
+int openconnect_webview_load_changed(struct openconnect_info *vpninfo,
+                                     const struct oc_webview_result *result);
+
+void openconnect_set_external_browser_callback(struct openconnect_info *vpninfo,
+					       openconnect_open_webview_vfn);
+
 /* Callback to allow binding a newly created socket's file descriptor to
    a specific interface, e.g. with SO_BINDTODEVICE. This tells the kernel
    not to route the traffic in question over the VPN tunnel. */
@@ -630,9 +804,9 @@ void openconnect_set_stats_handler(struct openconnect_info *vpninfo,
 int openconnect_has_pkcs11_support(void);
 
 /* The OpenSSL TPM ENGINE stores keys in a PEM file labelled with the string
-   -----BEGIN TSS KEY BLOB-----. GnuTLS may learn to support this format too,
-   in the near future. */
+   -----BEGIN TSS KEY BLOB-----. */
 int openconnect_has_tss_blob_support(void);
+int openconnect_has_tss2_blob_support(void);
 
 /* Software token capabilities. */
 int openconnect_has_stoken_support(void);
@@ -640,6 +814,10 @@ int openconnect_has_oath_support(void);
 int openconnect_has_yubioath_support(void);
 int openconnect_has_system_key_support(void);
 
+/* Query and select from among supported protocols */
+int openconnect_get_supported_protocols(struct oc_vpn_proto **protos);
+void openconnect_free_supported_protocols(struct oc_vpn_proto *protos);
+const char *openconnect_get_protocol(struct openconnect_info *vpninfo);
 int openconnect_set_protocol(struct openconnect_info *vpninfo, const char *protocol);
 
 struct addrinfo;
@@ -647,7 +825,7 @@ typedef int (*openconnect_getaddrinfo_vfn) (void *privdata, const char *node, co
 					    const struct addrinfo *hints, struct addrinfo **res);
 void openconnect_override_getaddrinfo(struct openconnect_info *vpninfo, openconnect_getaddrinfo_vfn gai_fn);
 
-/* Callback for configuring the interface after MTU detection finishes. */
+/* Callback for configuring the interface after tunnel is fully up. */
 typedef void (*openconnect_setup_tun_vfn) (void *privdata);
 void openconnect_set_setup_tun_handler(struct openconnect_info *vpninfo,
 				       openconnect_setup_tun_vfn setup_tun);
