@@ -511,7 +511,7 @@ int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
 		for (i = 0; i < sk_SSL_CIPHER_num(ciphers); i++) {
 			ssl_ciph = sk_SSL_CIPHER_value(ciphers, i);
 			/* For PSK-NEGOTIATE just use the first one we find */
-			if (!dtlsver || !strcmp(SSL_CIPHER_get_name(ssl_ciph), cipher))
+			if (!strcmp(SSL_CIPHER_get_name(ssl_ciph), cipher))
 				break;
 		}
 
@@ -539,9 +539,9 @@ int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("SSL_set_session() failed with DTLS protocol version 0x%x\n"
 				       "Are you using a version of OpenSSL older than 0.9.8m?\n"
-				       "See http://rt.openssl.org/Ticket/Display.html?id=1751\n"
+				       "See %s\n"
 				       "Use the --no-dtls command line option to avoid this message\n"),
-				     dtlsver);
+				     dtlsver, "http://rt.openssl.org/Ticket/Display.html?id=1751");
 			SSL_CTX_free(vpninfo->dtls_ctx);
 			SSL_free(dtls_ssl);
 			vpninfo->dtls_ctx = NULL;
