@@ -525,15 +525,25 @@ struct ossl_cert_info {
 	const char *certs_from;
 };
 
-static void free_ossl_cert_info(struct ossl_cert_info *oci)
+void unload_certificate(struct cert_info *certinfo, int finalize)
 {
-	if (oci->key)
-		EVP_PKEY_free(oci->key);
-	if (oci->cert)
-		X509_free(oci->cert);
-	if (oci->extra_certs)
-		sk_X509_pop_free(oci->extra_certs, X509_free);
-	memset(oci, 0, sizeof(*oci));
+	(void) finalize;
+
+	if (!certinfo)
+		return;
+
+	if (certinfo->priv_info) {
+		struct ossl_cert_info *oci = certinfo->priv_info;
+
+		certinfo->priv_info = NULL;
+		if (oci->key)
+			EVP_PKEY_free(oci->key);
+		if (oci->cert)
+			X509_free(oci->cert);
+		if (oci->extra_certs)
+			sk_X509_pop_free(oci->extra_certs, X509_free);
+		free(oci);
+	}
 }
 
 static int install_ssl_ctx_certs(struct openconnect_info *vpninfo, struct ossl_cert_info *oci)
@@ -674,7 +684,7 @@ static int load_pkcs12_certificate(struct openconnect_info *vpninfo, struct cert
 	PKCS12_free(p12);
 
 	if (ret)
-		free_ossl_cert_info(oci);
+		unload_certificate(certinfo, 1);
 
 	return ret;
 }
@@ -905,15 +915,13 @@ static int is_pem_password_error(struct openconnect_info *vpninfo, struct cert_i
 	return 0;
 }
 
-static int load_certificate(struct openconnect_info *vpninfo, struct cert_info *certinfo,
+static int xload_certificate(struct openconnect_info *vpninfo, struct cert_info *certinfo,
 			    struct ossl_cert_info *oci)
 {
 	FILE *f;
 	char buf[256];
 	int ret;
 
-	certinfo->vpninfo = vpninfo;
-
 	if (!strncmp(certinfo->cert, "pkcs11:", 7)) {
 		int ret = load_pkcs11_certificate(vpninfo, certinfo, &oci->cert);
 		if (ret)
@@ -1124,6 +1132,29 @@ static int load_certificate(struct openconnect_info *vpninfo, struct cert_info *
 	return -EINVAL;
 }
 
+int load_certificate(struct openconnect_info *vpninfo, struct cert_info *certinfo, int flags)
+{
+	struct ossl_cert_info *oci;
+	int ret;
+
+	(void) flags;
+
+	certinfo->priv_info = oci = calloc(1, sizeof(*oci));
+	if (!oci) {
+		ret = -ENOMEM;
+		goto done;
+	}
+	certinfo->vpninfo = vpninfo;
+
+	ret = xload_certificate(vpninfo, certinfo, oci);
+
+done:
+	if (ret)
+		unload_certificate(certinfo, 1);
+
+	return ret;
+}
+
 static int get_cert_fingerprint(struct openconnect_info *vpninfo,
 				X509 *cert, const EVP_MD *type,
 				char *buf)
@@ -1698,11 +1729,13 @@ static int check_certificate_expiry(struct openconnect_info *vpninfo, struct cer
 
 static int load_primary_certificate(struct openconnect_info *vpninfo)
 {
-	struct ossl_cert_info oci = { };
+	struct cert_info *certinfo = &vpninfo->certinfo[0];
+	struct ossl_cert_info *oci;
 
-	int ret = load_certificate(vpninfo, &vpninfo->certinfo[0], &oci);
+	int ret = load_certificate(vpninfo, certinfo, 0);
+	oci = certinfo->priv_info;
 	if (!ret)
-		ret = install_ssl_ctx_certs(vpninfo, &oci);
+		ret = install_ssl_ctx_certs(vpninfo, oci);
 
 	if (!ret && !SSL_CTX_check_private_key(vpninfo->https_ctx)) {
 		vpn_progress(vpninfo, PRG_ERR,
@@ -1710,9 +1743,9 @@ static int load_primary_certificate(struct openconnect_info *vpninfo)
 		ret = -EINVAL;
 	}
 	if (!ret)
-		check_certificate_expiry(vpninfo, &vpninfo->certinfo[0], &oci);
+		check_certificate_expiry(vpninfo, &vpninfo->certinfo[0], oci);
 
-	free_ossl_cert_info(&oci);
+	unload_certificate(certinfo, 1);
 	return ret;
 }
 
@@ -1868,13 +1901,10 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 			struct oc_text_buf *buf = buf_alloc();
 			if (vpninfo->pfs)
 				buf_append(buf, "HIGH:!aNULL:!eNULL:-RSA");
+			else if (vpninfo->allow_insecure_crypto)
+				buf_append(buf, "ALL");
 			else
-				buf_append(buf, "DEFAULT");
-
-			if (vpninfo->allow_insecure_crypto)
-				buf_append(buf, ":+3DES:+RC4");
-			else
-				buf_append(buf, ":-3DES:-RC4");
+				buf_append(buf, "DEFAULT:-3DES:-RC4");
 
 			if (buf_error(buf)) {
 				vpn_progress(vpninfo, PRG_ERR,
@@ -2247,3 +2277,432 @@ void destroy_eap_ttls(struct openconnect_info *vpninfo, void *ttls)
 	/* Leave the BIO_METH for now. It may get reused and we don't want to
 	 * have to call BIO_get_new_index() more times than is necessary */
 }
+
+#ifdef HAVE_HPKE_SUPPORT
+static int generate_strap_key(EC_KEY **key, char **pubkey,
+			      unsigned char *privder_in, int privderlen,
+			      unsigned char **pubder, int *pubderlen)
+{
+	EC_KEY *lkey;
+	struct oc_text_buf *buf = NULL;
+	unsigned char *der = NULL;
+	int len;
+
+	if (privder_in) {
+		lkey = d2i_ECPrivateKey(NULL, (const unsigned char **)&privder_in, privderlen);
+		if (!lkey)
+			return -EIO;
+	} else {
+		lkey = EC_KEY_new_by_curve_name(NID_X9_62_prime256v1);
+		if (!lkey)
+			return -EIO;
+
+		if (!EC_KEY_generate_key(lkey)) {
+			EC_KEY_free(lkey);
+			return -EIO;
+		}
+	}
+
+	len = i2d_EC_PUBKEY(lkey, &der);
+	buf = buf_alloc();
+	buf_append_base64(buf, der, len, 0);
+	if (buf_error(buf)) {
+		EC_KEY_free(lkey);
+		free(der);
+		return buf_free(buf);
+	}
+
+	/* All done. There are no failure modes from here on, so
+	 * install the resulting key/pubkey/etc. where the caller
+	 * asked us to, freeing the previous ones if needed. */
+	EC_KEY_free(*key);
+	*key = lkey;
+
+	free(*pubkey);
+	*pubkey = buf->data;
+
+	/* If the caller wants the DER, give it to them */
+	if (pubder && pubderlen) {
+		*pubder = der;
+		*pubderlen = len;
+	} else {
+		free(der);
+	}
+
+	buf->data = NULL;
+	buf_free(buf);
+	return 0;
+}
+
+int generate_strap_keys(struct openconnect_info *vpninfo)
+{
+	int err;
+
+	err = generate_strap_key(&vpninfo->strap_key, &vpninfo->strap_pubkey,
+				 NULL, 0, NULL, NULL);
+	if (err) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to generate STRAP key"));
+		openconnect_report_ssl_errors(vpninfo);
+		free_strap_keys(vpninfo);
+		return -EIO;
+	}
+
+	err = generate_strap_key(&vpninfo->strap_dh_key, &vpninfo->strap_dh_pubkey,
+				 NULL, 0, NULL, NULL);
+	if (err) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to generate STRAP DH key\n"));
+		openconnect_report_ssl_errors(vpninfo);
+		free_strap_keys(vpninfo);
+		return -EIO;
+	}
+	return 0;
+}
+
+void free_strap_keys(struct openconnect_info *vpninfo)
+{
+	if (vpninfo->strap_key)
+		EC_KEY_free(vpninfo->strap_key);
+	if (vpninfo->strap_dh_key)
+		EC_KEY_free(vpninfo->strap_dh_key);
+
+	vpninfo->strap_key = vpninfo->strap_dh_key = NULL;
+}
+
+int ingest_strap_privkey(struct openconnect_info *vpninfo, unsigned char *der, int len)
+{
+	if (generate_strap_key(&vpninfo->strap_key,
+			       &vpninfo->strap_pubkey, der, len, NULL, 0)) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to decode STRAP key\n"));
+		openconnect_report_ssl_errors(vpninfo);
+		return -EIO;
+	}
+	return 0;
+}
+
+void append_strap_privkey(struct openconnect_info *vpninfo,
+			  struct oc_text_buf *buf)
+{
+	unsigned char *der = NULL;
+	int derlen = i2d_ECPrivateKey(vpninfo->strap_key, &der);
+	if (derlen > 0)
+		buf_append_base64(buf, der, derlen, 0);
+}
+
+#include <openssl/kdf.h>
+
+int ecdh_compute_secp256r1(struct openconnect_info *vpninfo, const unsigned char *pubkey,
+			   int pubkey_len, unsigned char *secret)
+{
+	const EC_POINT *point;
+	EC_KEY *pkey;
+	int ret = 0;
+
+	if (!(pkey = d2i_EC_PUBKEY(NULL, &pubkey, pubkey_len)) ||
+	    !(point = EC_KEY_get0_public_key(pkey))) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to decode server DH key\n"));
+		openconnect_report_ssl_errors(vpninfo);
+		ret = -EIO;
+		goto out;
+
+	}
+
+	/* Perform the DH secret derivation from our STRAP-DH key
+	 * and the one the server returned to us in the payload. */
+	if (ECDH_compute_key(secret, 32, point, vpninfo->strap_dh_key, NULL) <= 0) {
+		vpn_progress(vpninfo, PRG_ERR, _("Failed to compute DH secret\n"));
+		openconnect_report_ssl_errors(vpninfo);
+		ret = -EIO;
+	}
+ out:
+	EC_KEY_free(pkey);
+	return ret;
+}
+
+int hkdf_sha256_extract_expand(struct openconnect_info *vpninfo, unsigned char *buf,
+			       const char *info, int infolen)
+{
+	size_t buflen = 32;
+	int ret = 0;
+
+	/* Next, use HKDF to generate the actual key used for encryption. */
+	EVP_PKEY_CTX *ctx = EVP_PKEY_CTX_new_id(EVP_PKEY_HKDF, NULL);
+	if (!ctx || !EVP_PKEY_derive_init(ctx) ||
+	    !EVP_PKEY_CTX_set_hkdf_md(ctx, EVP_sha256()) ||
+	    !EVP_PKEY_CTX_set1_hkdf_key(ctx, buf, buflen) ||
+	    !EVP_PKEY_CTX_hkdf_mode(ctx, EVP_PKEY_HKDEF_MODE_EXTRACT_AND_EXPAND) ||
+	    !EVP_PKEY_CTX_add1_hkdf_info(ctx, info, infolen) ||
+	    EVP_PKEY_derive(ctx, buf, &buflen) != 1) {
+		vpn_progress(vpninfo, PRG_ERR, _("HKDF key derivation failed\n"));
+		openconnect_report_ssl_errors(vpninfo);
+		ret = -EINVAL;
+	}
+	EVP_PKEY_CTX_free(ctx);
+	return ret;
+}
+
+int aes_256_gcm_decrypt(struct openconnect_info *vpninfo, unsigned char *key,
+			unsigned char *data, int len,
+			unsigned char *iv, unsigned char *tag)
+{
+	/* Finally, we actually decrypt the sso-token */
+	EVP_CIPHER_CTX *cctx = EVP_CIPHER_CTX_new();
+	int ret = 0, i = 0;
+
+	if (!cctx ||
+	    !EVP_DecryptInit_ex(cctx, EVP_aes_256_gcm(), NULL, key, iv) ||
+	    !EVP_CIPHER_CTX_ctrl(cctx, EVP_CTRL_AEAD_SET_TAG, 12, tag) ||
+	    !EVP_DecryptUpdate(cctx, data, &len, data, len) ||
+	    !EVP_DecryptFinal(cctx, NULL, &i)) {
+		vpn_progress(vpninfo, PRG_ERR, _("SSO token decryption failed\n"));
+		openconnect_report_ssl_errors(vpninfo);
+		ret = -EINVAL;
+	}
+	EVP_CIPHER_CTX_free(cctx);
+	return ret;
+}
+
+void append_strap_verify(struct openconnect_info *vpninfo,
+			 struct oc_text_buf *buf, int rekey)
+{
+	unsigned char finished[64];
+	size_t flen = SSL_get_finished(vpninfo->https_ssl, finished, sizeof(finished));
+
+	if (flen > sizeof(finished)) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("SSL Finished message too large (%zd bytes)\n"), flen);
+		if (!buf_error(buf))
+			buf->error = -EIO;
+		return;
+	}
+
+	/* If we're rekeying, we need to sign the Verify header with the *old* key. */
+	EVP_PKEY *evpkey = EVP_PKEY_new();
+	if (!evpkey || EVP_PKEY_set1_EC_KEY(evpkey, vpninfo->strap_key) <= 0) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("STRAP signature failed\n"));
+	fail_errors:
+		openconnect_report_ssl_errors(vpninfo);
+	fail_pkey:
+		if (!buf_error(buf))
+			buf->error = -EIO;
+		EVP_PKEY_free(evpkey);
+		return;
+	}
+
+	unsigned char *pubkey_der = NULL;
+	int pubkey_derlen = 0;
+	if (rekey) {
+		if (generate_strap_key(&vpninfo->strap_key, &vpninfo->strap_pubkey,
+				       NULL, 0, &pubkey_der, &pubkey_derlen)) {
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Failed to regenerate STRAP key\n"));
+			goto fail_errors;
+		}
+	} else {
+		pubkey_der = openconnect_base64_decode(&pubkey_derlen, vpninfo->strap_pubkey);
+		if (!pubkey_der) {
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Failed to generate STRAP key DER\n"));
+			goto fail_pkey;
+		}
+	}
+
+	EVP_MD_CTX *mdctx = EVP_MD_CTX_new();
+	const EVP_MD *md = EVP_sha256(); /* We only support prime256v1 for now */
+
+	unsigned char signature_bin[128];
+	size_t siglen = sizeof(signature_bin);
+	int ok = (mdctx &&
+		  EVP_DigestSignInit(mdctx, NULL, md, NULL, evpkey) > 0 &&
+		  EVP_DigestSignUpdate(mdctx, finished, flen) > 0 &&
+		  EVP_DigestSignUpdate(mdctx, pubkey_der, pubkey_derlen) > 0 &&
+		  EVP_DigestSignFinal(mdctx, (void *)signature_bin, &siglen) > 0);
+
+	EVP_MD_CTX_free(mdctx);
+	EVP_PKEY_free(evpkey);
+	free(pubkey_der);
+
+	if (!ok) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("STRAP signature failed\n"));
+		goto fail_errors;
+	}
+
+	buf_append_base64(buf, signature_bin, siglen, 0);
+}
+#endif /* HAVE_HPKE_SUPPORT */
+
+int export_certificate_pkcs7(struct openconnect_info *vpninfo,
+			     struct cert_info *certinfo,
+			     cert_format_t format,
+			     struct oc_text_buf **pp7b)
+{
+	struct ossl_cert_info *oci;
+	PKCS7 *p7 = NULL;
+	BIO *bio = NULL;
+	BUF_MEM *bptr = NULL;
+	struct oc_text_buf *p7b = NULL;
+	int ret, ok;
+
+	if (!(certinfo && (oci = certinfo->priv_info) && pp7b))
+		return -EINVAL;
+
+	/* We have the client certificate in 'oci.cert' and *optionally*
+	 * a stack of intermediate certs in oci.extra_certs. For the TLS
+	 * connection those would be used by SSL_CTX_use_certificate() and
+	 * SSL_CTX_add_extra_chain_cert() respectively. For PKCS7_sign()
+	 * we need the actual cert at the head of the stack, so *create*
+	 * one if needed, and insert oci.cert at position zero. */
+
+	if (!oci->extra_certs)
+		oci->extra_certs = sk_X509_new_null();
+	if (!oci->extra_certs)
+		goto err;
+	if (!sk_X509_insert(oci->extra_certs, oci->cert, 0))
+		goto err;
+	X509_up_ref(oci->cert);
+
+	p7 = PKCS7_sign(NULL, NULL, oci->extra_certs, NULL, PKCS7_DETACHED);
+	if (!p7) {
+	err:
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to create PKCS#7 structure\n"));
+		ret = -EIO;
+		goto out;
+	}
+
+	ret = 0;
+
+	bio = BIO_new(BIO_s_mem());
+	if (!bio) {
+		ret = -ENOMEM;
+		goto pkcs7_error;
+	}
+
+	if (format == CERT_FORMAT_ASN1) {
+		ok = i2d_PKCS7_bio(bio, p7);
+	} else if (format == CERT_FORMAT_PEM) {
+		ok = PEM_write_bio_PKCS7(bio, p7);
+	} else {
+		ret = -EINVAL;
+		goto pkcs7_error;
+	}
+
+	if (!ok) {
+		ret = -EIO;
+		goto pkcs7_error;
+	}
+
+	BIO_get_mem_ptr(bio, &bptr);
+
+	p7b = buf_alloc();
+	if (!p7b)
+		ret = -ENOMEM;
+
+	if (ret < 0) {
+pkcs7_error:
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to output PKCS#7 structure\n"));
+		goto out;
+	}
+
+	BIO_set_close(bio, BIO_NOCLOSE);
+
+	p7b->data = bptr->data;
+	p7b->pos = bptr->length;
+
+	*pp7b = p7b;
+	p7b = NULL;
+
+out:
+	buf_free(p7b);
+	BIO_free(bio);
+	if (p7)
+		PKCS7_free(p7);
+	return ret;
+}
+
+int multicert_sign_data(struct openconnect_info *vpninfo,
+			struct cert_info *certinfo,
+			unsigned int hashes,
+			const void *chdata, size_t chdata_len,
+			struct oc_text_buf **psignature)
+{
+	struct table_entry {
+		openconnect_hash_type id;
+		const EVP_MD *(*evp_md_fn)(void);
+	};
+	static struct table_entry table[] = {
+		{ OPENCONNECT_HASH_SHA512, &EVP_sha512 },
+		{ OPENCONNECT_HASH_SHA384, &EVP_sha384 },
+		{ OPENCONNECT_HASH_SHA256, &EVP_sha256 },
+		{ OPENCONNECT_HASH_UNKNOWN },
+	};
+	struct ossl_cert_info *oci;
+	struct oc_text_buf *signature;
+	openconnect_hash_type hash;
+	int ret;
+
+	/**
+	 * Check preconditions...
+	 */
+	if (!(certinfo && (oci = certinfo->priv_info)
+	      && hashes && chdata && chdata_len && psignature))
+		return -EINVAL;
+
+	*psignature = NULL;
+
+	signature = buf_alloc();
+	if (!signature)
+		goto out_of_memory;
+
+	for (const struct table_entry *entry = table;
+	     (hash = entry->id) != OPENCONNECT_HASH_UNKNOWN;
+	     entry++) {
+		if ((hashes & MULTICERT_HASH_FLAG(hash)) == 0)
+			continue;
+
+		EVP_MD_CTX *mdctx = EVP_MD_CTX_new();
+		if (!mdctx)
+			goto out_of_memory;
+
+		const EVP_MD *md = (*entry->evp_md_fn)();
+
+		size_t siglen = 0;
+		int ok = (EVP_DigestSignInit(mdctx, NULL, md, NULL, oci->key) > 0 &&
+			  EVP_DigestSignUpdate(mdctx, chdata, chdata_len) > 0 &&
+			  EVP_DigestSignFinal(mdctx, NULL, &siglen) > 0 &&
+			  !buf_ensure_space(signature, siglen) &&
+			  EVP_DigestSignFinal(mdctx, (void *)signature->data, &siglen) > 0);
+
+		EVP_MD_CTX_free(mdctx);
+
+		if (ok) {
+			signature->pos = siglen;
+			*psignature = signature;
+			return hash;
+		}
+	}
+
+	/** Error path */
+
+	ret = -EIO;
+
+	if (buf_error(signature)) {
+out_of_memory:
+		ret = -ENOMEM;
+	}
+
+	buf_free(signature);
+
+	vpn_progress(vpninfo, PRG_ERR,
+		     _("Failed to generate signature for multiple certificate authentication\n"));
+	openconnect_report_ssl_errors(vpninfo);
+
+	return ret;
+}
+
