@@ -157,7 +157,7 @@ static int unhdlc_in_place(struct openconnect_info *vpninfo, unsigned char *byte
 	}
 }
 
-static const char *ppps_names[] = {
+static const char * const ppps_names[] = {
 	"DEAD",
 	"ESTABLISH",
 	"OPENED",
@@ -166,7 +166,7 @@ static const char *ppps_names[] = {
 	"TERMINATE"
 };
 
-static const char *encap_names[PPP_ENCAP_MAX+1] = {
+static const char * const encap_names[PPP_ENCAP_MAX+1] = {
 	NULL,
 	"RFC1661",
 	"RFC1662 HDLC",
@@ -175,7 +175,7 @@ static const char *encap_names[PPP_ENCAP_MAX+1] = {
 	"FORTINET",
 };
 
-static const char *lcp_names[] = {
+static const char * const lcp_names[] = {
 	NULL,
 	"Configure-Request",
 	"Configure-Ack",
