@@ -17,11 +17,10 @@
 
 #include <config.h>
 
-#include <errno.h>
-#include <string.h>
-
 #include "openconnect-internal.h"
 
+#include <errno.h>
+#include <string.h>
 
 static int sspi_setup(struct openconnect_info *vpninfo, struct http_auth_state *auth_state, const char *service, int proxy)
 {
@@ -120,7 +119,7 @@ int gssapi_authorization(struct openconnect_info *vpninfo, int proxy,
 	}
 
 	buf_append(hdrbuf, "%sAuthorization: Negotiate ", proxy ? "Proxy-" : "");
-	buf_append_base64(hdrbuf, out_token.pvBuffer, out_token.cbBuffer);
+	buf_append_base64(hdrbuf, out_token.pvBuffer, out_token.cbBuffer, 0);
 	buf_append(hdrbuf, "\r\n");
 
 	FreeContextBuffer(out_token.pvBuffer);
