@@ -186,7 +186,6 @@ intptr_t os_setup_tun(struct openconnect_info *vpninfo)
 
 #else /* !__sun__ && !__native_client__ */
 
-/* MTU setting code for both Linux and BSD systems */
 static void ifreq_set_ifname(struct openconnect_info *vpninfo, struct ifreq *ifr)
 {
 	char *ifname = openconnect_utf8_to_legacy(vpninfo, vpninfo->ifname);
@@ -195,28 +194,6 @@ static void ifreq_set_ifname(struct openconnect_info *vpninfo, struct ifreq *ifr
 		free(ifname);
 }
 
-static int set_tun_mtu(struct openconnect_info *vpninfo)
-{
-	struct ifreq ifr;
-	int net_fd;
-
-	net_fd = socket(PF_INET, SOCK_DGRAM, 0);
-	if (net_fd < 0) {
-		vpn_perror(vpninfo, _("open net"));
-		return -EINVAL;
-	}
-
-	memset(&ifr, 0, sizeof(ifr));
-	ifreq_set_ifname(vpninfo, &ifr);
-	ifr.ifr_mtu = vpninfo->ip_info.mtu;
-
-	if (ioctl(net_fd, SIOCSIFMTU, &ifr) < 0)
-		vpn_perror(vpninfo, _("SIOCSIFMTU"));
-
-	close(net_fd);
-	return 0;
-}
-
 #ifdef IFF_TUN /* Linux */
 intptr_t os_setup_tun(struct openconnect_info *vpninfo)
 {
@@ -255,7 +232,8 @@ intptr_t os_setup_tun(struct openconnect_info *vpninfo)
 		if (err == EPERM) {
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("To configure local networking, openconnect must be running as root\n"
-				       "See https://www.infradead.org/openconnect/nonroot.html for more information\n"));
+				       "See %s for more information\n"),
+				     "https://www.infradead.org/openconnect/nonroot.html");
 		}
 		close(tun_fd);
 		return -EIO;
@@ -263,9 +241,6 @@ intptr_t os_setup_tun(struct openconnect_info *vpninfo)
 	if (!vpninfo->ifname)
 		vpninfo->ifname = strdup(ifr.ifr_name);
 
-	/* Ancient vpnc-scripts might not get this right */
-	set_tun_mtu(vpninfo);
-
 	return tun_fd;
 }
 #else /* BSD et al, including OS X */
@@ -351,7 +326,7 @@ intptr_t os_setup_tun(struct openconnect_info *vpninfo)
 	do {
 		sc.sc_unit = unit_nr + 1;
 
-		if (!connect(tun_fd, (struct sockaddr * )&sc, sizeof(sc))) {
+		if (!connect(tun_fd, (struct sockaddr *)&sc, sizeof(sc))) {
 			if (!vpninfo->ifname &&
 			    asprintf(&vpninfo->ifname, "utun%d", unit_nr) == -1) {
 				vpn_progress(vpninfo, PRG_ERR,
@@ -393,10 +368,9 @@ intptr_t os_setup_tun(struct openconnect_info *vpninfo)
 			 "/dev/%s", vpninfo->ifname);
 		tun_fd = bsd_open_tun(tun_name);
 		if (tun_fd < 0) {
-			int err = errno;
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("Cannot open '%s': %s\n"),
-				     tun_name, strerror(err));
+				     tun_name, strerror(errno));
 			return -EINVAL;
 		}
 	}
@@ -440,9 +414,6 @@ intptr_t os_setup_tun(struct openconnect_info *vpninfo)
 	}
 #endif
 
-	/* Ancient vpnc-scripts might not get this right */
-	set_tun_mtu(vpninfo);
-
 	return tun_fd;
 }
 #endif /* !IFF_TUN (i.e. BSD) */
