@@ -17,13 +17,13 @@
 
 #include <config.h>
 
+#include "openconnect-internal.h"
+
 #include <ctype.h>
 #include <errno.h>
 #include <stdlib.h>
 #include <string.h>
 
-#include "openconnect-internal.h"
-
 #define NAME_TAG	0x71
 #define NAME_LIST_TAG	0x72
 #define KEY_TAG		0x73
@@ -55,7 +55,7 @@ static const unsigned char send_remaining[] = { 0x00, SEND_REMAINING_INS, 0x00,
 #define free_scard_error(str) free(str)
 #else
 #define scard_error(st) ((char *)pcsc_stringify_error(st))
-#define free_scard_error(str) do { ; } while (0)
+#define free_scard_error(str)
 
 #endif
 
@@ -151,7 +151,7 @@ static int buf_tlv(struct oc_text_buf *buf, int *loc, unsigned char *type)
 			return -EINVAL;
 		len = (unsigned char)buf->data[(*loc)++];
 		len <<= 8;
-		len = (unsigned char)buf->data[(*loc)++];
+		len |= (unsigned char)buf->data[(*loc)++];
 		left -= 2;
 	}
 
@@ -212,7 +212,7 @@ static int select_yubioath_applet(struct openconnect_info *vpninfo,
 		chall_len = tlvlen;
 
 		/* On later invocations, we know there must have been a
-		 * successfui authentication in the past. So try the same
+		 * successful authentication in the past. So try the same
 		 * hash first, and only retry in the loop on failure. */
 		if (!vpninfo->pcsc) {
 			struct oc_auth_form f;
@@ -277,8 +277,8 @@ static int select_yubioath_applet(struct openconnect_info *vpninfo,
 			memset(vpninfo->yubikey_pwhash, 0, sizeof(vpninfo->yubikey_pwhash));
 			if (pin) {
 				/* Try working around pre-KitKat PBKDF2 bug discussed at
-				 * http://forum.yubico.com/viewtopic.php?f=26&t=1601#p6807 and
-				 * http://android-developers.blogspot.se/2013/12/changes-to-secretkeyfactory-api-in.html */
+				 * https://forum.yubico.com/viewtopica454-3.html?f=26&t=1601#p6807 and
+				 * https://android-developers.googleblog.com/2013/12/changes-to-secretkeyfactory-api-in.html */
 				const char *in;
 				char *out;
 
